DATETIME_COMPONENTS_PREFIX = (
    "y",  # year
    "q",  # quarter
    "m",  # month
    "w",  # week
    "d",  # day of week
    "h",  # hour
    "mm",  # minute
    "s",  # second
)

PREDICTION_COLNAME = "prediction"
